#ifndef _PANEL_H
extern "C" {
#include <panel.h>
}
#define _PANEL_H
#endif
#ifndef _WIN32
#include <unistd.h> // write
#else
#include <windows.h>
#include <io.h>
#endif
#include <ctype.h>

#include "AUI.h"
#include "Hero.h"

int ColorMap[22];
const char *splash_logo[] =
{ /* PRIME logo designed by Psiweapon. */
    ".   _______________________   .",
    "  _/                       \\_  ",
    "=|  ####  ####  # #   #####  |=",
    " >------#-----#-#-##-##------< ",
    "=|  ####  ####  # # # #####  |=",
    "=|  #     #   # # #   #      |=",
    "=|_ #     #   # # #   ##### _|=",
    ".  \\[                     ]/  ."
};

shASCIIInterface::shASCIIInterface ()
{
    WINDOW *win = initscr ();

    if (!win) {
        fprintf (stderr, "Sorry, curses support is required.\n");
        exitPRIME (-1);
    }

    if (!has_colors ()) {
        endwin ();
        fprintf (stderr, "Sorry, color support is required.\n");
        exitPRIME (-1);
    }
    start_color ();

    for (int i = 1; i < COLOR_PAIRS; ++i) {
        init_pair (i, i %8, i/8);
    }


#ifdef _WIN32
    curs_set (1);
    ColorMap[kBlack] = COLOR_BLACK;
    ColorMap[kRed] = COLOR_PAIR (4);
    ColorMap[kGreen] = COLOR_PAIR (2);
    ColorMap[kYellow] = COLOR_PAIR (6);
    ColorMap[kBlue] = COLOR_PAIR (1);
    ColorMap[kMagenta] = COLOR_PAIR (5);
    ColorMap[kCyan] = COLOR_PAIR (3);
    ColorMap[kGray] = COLOR_PAIR (7);
    ColorMap[kBrightRed] = COLOR_PAIR (4) | A_BOLD;
    ColorMap[kBrightGreen] = COLOR_PAIR (2) | A_BOLD;
    ColorMap[kBrightYellow] = COLOR_PAIR (6) | A_BOLD;
    ColorMap[kBrightBlue] = COLOR_PAIR (1)| A_BOLD;
    ColorMap[kBrightMagenta] = COLOR_PAIR (5) | A_BOLD;
    ColorMap[kBrightCyan] = COLOR_PAIR (3) | A_BOLD;
    ColorMap[kWhite] = COLOR_PAIR (7) | A_BOLD;

    ColorMap[kDarkRed] = COLOR_PAIR (4) | A_DIM;
    ColorMap[kDarkGreen] = COLOR_PAIR (2) | A_DIM;
    ColorMap[kBrown] = COLOR_PAIR (6);
    ColorMap[kDarkBlue] = COLOR_PAIR (1)| A_DIM;
    ColorMap[kDarkMagenta] = COLOR_PAIR (5) | A_DIM;
    ColorMap[kDarkCyan] = COLOR_PAIR (3) | A_DIM;
    ColorMap[kDarkGray] = COLOR_PAIR (7) | A_DIM;

#else
    ColorMap[kBlack] = COLOR_BLACK;
    ColorMap[kRed] = COLOR_PAIR (1);
    ColorMap[kGreen] = COLOR_PAIR (2);
    ColorMap[kYellow] = COLOR_PAIR (3);
    ColorMap[kBlue] = COLOR_PAIR (4);
    ColorMap[kMagenta] = COLOR_PAIR (5);
    ColorMap[kCyan] = COLOR_PAIR (6);
    ColorMap[kGray] = COLOR_PAIR (7);
    ColorMap[kBrightRed] = COLOR_PAIR (1) | A_BOLD;
    ColorMap[kBrightGreen] = COLOR_PAIR (2) | A_BOLD;
    ColorMap[kBrightYellow] = COLOR_PAIR (3) | A_BOLD;
    ColorMap[kBrightBlue] = COLOR_PAIR (4)| A_BOLD;
    ColorMap[kBrightMagenta] = COLOR_PAIR (5) | A_BOLD;
    ColorMap[kBrightCyan] = COLOR_PAIR (6) | A_BOLD;
    ColorMap[kWhite] = COLOR_PAIR (7) | A_BOLD;

    /* A_DIM doesn't seem to do much, but... */
    ColorMap[kDarkRed] = COLOR_PAIR (1) | A_DIM;
    ColorMap[kDarkGreen] = COLOR_PAIR (2) | A_DIM;
    ColorMap[kBrown] = COLOR_PAIR (3) | A_DIM;
    ColorMap[kDarkBlue] = COLOR_PAIR (4)| A_DIM;
    ColorMap[kDarkMagenta] = COLOR_PAIR (5) | A_DIM;
    ColorMap[kDarkCyan] = COLOR_PAIR (6) | A_DIM;
    ColorMap[kDarkGray] = COLOR_PAIR (7) | A_DIM;


#endif

    mColor = kGray;

//    cbreak ();
    raw ();
    noecho ();
#ifdef DJGPP
    nl ();
#else
    nonl ();
#endif

    mXMax = 64;
    mYMax = 20;

    mMainWin = newwin (20, mXMax, 0, 0);
    if (!mMainWin) goto toosmall;
    new_panel (mMainWin);
    notimeout (mMainWin, TRUE);
    mSideWin = newwin (20, 80 - mXMax, 0, mXMax);
    if (!mSideWin) goto toosmall;
    new_panel (mSideWin);
    mLogWin = newwin (5, 80, 20, 0);
    if (!mLogWin) goto toosmall;
    mDiagWin = NULL;
    if (GodMode && LINES >= 30) {
        mDiagWin = newwin (0, 80, 25, 0);
        scrollok (mDiagWin, TRUE);
    }
    notimeout (mLogWin, TRUE);
    mLogSize = 5;
    mLogRow = 0;
    mLogCol = 0;
    mHistoryIdx = 0;
    mHistoryWrapped = 0;
    mNoNewline = 0;
    mLogSCount = 0;
    mPause = 0;

    mX0 = 0;
    mY0 = 0;

    intrflush (mMainWin, FALSE);
    keypad (mMainWin, TRUE);
    intrflush (mLogWin, FALSE);
    scrollok (mLogWin, TRUE);
#ifndef WIN32
    set_escdelay (50);
#endif

    initializeGlyphs ();

    debug.log ("COLORS: %d", COLORS);
    debug.log ("COLOR_PAIRS: %d", COLOR_PAIRS);
    debug.log ("can change: %d", can_change_color());

    return;

toosmall:
    endwin ();
    fprintf (stderr, "Sorry, but a terminal with dimensions of at least 80x25"
             " is required.\n");
    exitPRIME (-1);
}

shASCIIInterface::~shASCIIInterface ()
{
    if (mMainWin) endwin ();
}


void
shASCIIInterface::cursorOnXY (int x, int y)
{
    wmove (mMainWin, y, x);
    wrefresh (mMainWin); /* Necessary for cursor to be actually moved. */
}


void
shASCIIInterface::outChar (char ch)
{
    addch (ch);
}


void
shASCIIInterface::outStr (const char *str)
{
    addstr (str);
}


void
shASCIIInterface::assign (int key, Command cmd)
{
    mKey2Cmd[key] = cmd;
}


void
shASCIIInterface::resetKeybindings ()
{
    for (int i = 0; i < KEY_MAX; i++) {
        mKey2Cmd[i] = kNoCommand;
    }
}


void
shASCIIInterface::readKeybindings (const char *fname)
{
    static const struct shNameToKey KeyNames[] =
    {
        {"break", KEY_BREAK},
        {"space", ' '},
        {"tab", '\t'},
        {"backspace", 8},
        {"enter", KEY_ENTER},
        {"down arrow", KEY_DOWN},
        {"up arrow", KEY_UP},
        {"left arrow", KEY_LEFT},
        {"right arrow", KEY_RIGHT},
        {"keypad 1", KEY_C1},
        {"keypad 2", KEY_DOWN},
        {"keypad 3", KEY_C3},
        {"keypad 4", KEY_LEFT},
        {"keypad 5", KEY_B2},
        {"keypad 6", KEY_RIGHT},
        {"keypad 7", KEY_A1},
        {"keypad 8", KEY_UP},
        {"keypad 9", KEY_A3},
        {"home", KEY_HOME},
        {"end", KEY_END},
        {"insert", KEY_IC},
        {"delete", KEY_DC},
        {"page up", KEY_PPAGE},
        {"page down", KEY_NPAGE},
        {"shift home", KEY_SHOME},
        {"shift end", KEY_SEND},
        {"shift insert", KEY_SIC},
        {"shift delete", KEY_SDC},
        {"fn 1", KEY_F(1)},
        {"fn 2", KEY_F(2)},
        {"fn 3", KEY_F(3)},
        {"fn 4", KEY_F(4)},
        {"fn 5", KEY_F(5)},
        {"fn 6", KEY_F(6)},
        {"fn 7", KEY_F(7)},
        {"fn 8", KEY_F(8)},
        {"fn 9", KEY_F(9)},
        {"fn 10", KEY_F(10)},
        {"fn 11", KEY_F(11)},
        {"fn 12", KEY_F(12)}
    };
    const int n_keys = sizeof (KeyNames) / sizeof (shNameToKey);
    I->readKeybindings (fname, n_keys, KeyNames);
}


const char *
shASCIIInterface::getKeyForCommand (Command cmd)
{   /* Find key corresponding to command. */
    int i;
    for (i = 0; i < KEY_MAX && cmd != mKey2Cmd[i]; ++i)
        ;
    if (i == KEY_MAX) /* Unassigned? This is bad. */
        return NULL;  /* Safe, will not lead to crash. */
    char *buf = GetBuf ();
    if (i == '\t') {
        sprintf (buf, "tab");
    } else if (i == ' ') {
        sprintf (buf, "space");
    } else if (i < ' ') {
        sprintf (buf, "ctrl %c", i+64);
    } else if (i >= KEY_F(1) && i <= KEY_F(12)) {
        sprintf (buf, "F%d", i - KEY_F(1) + 1);
    } else {
        sprintf (buf, "%c", i);
    }
    return buf;
}


shInterface::Command
shASCIIInterface::keyToCommand (int key)
{
    return mKey2Cmd[key];
}


void
shASCIIInterface::moreBegin ()
{
    static const char     more[] = "  --More--";
    mvwaddstr (mLogWin, mLogRow - 1, mLogCol, more);
    wrefresh (mLogWin); /* Show cursor after more. */
    drawLog ();
}


void
shASCIIInterface::moreEnd ()
{
    static const char antimore[] = "          ";
    if (Flags.mFadeLog) {
        mvwaddstr (mLogWin, mLogRow - 1, mLogCol, antimore);
    }
    pageLog ();
}


int
shASCIIInterface::getChar ()
{
    return wgetch (mMainWin);
}


/* MODIFIES: reads a string from the keyboard
   RETURNS: number of characters read, 0 on error */

int
shASCIIInterface::getStr (char *buf, int len, const char *prompt,
                    const char *dflt /* = NULL */ ) /* Default suggestion. */
{   /* can't use wgetnstr b/c it won't pass us ^C */
    char msg[80];
    int pos = 0;
    int savehistidx = mHistoryIdx;

    snprintf (msg, 80, "%s ", prompt);
    msg[79] = 0;
    p (msg);
    buf[0] = 0;

    if (dflt) {
        strncpy (buf, dflt, len);
        buf[len-1] = 0;
        waddstr (mLogWin, buf);
        pos = strlen (buf);
    }

    while (1) {
        drawLog ();
        int c = getChar ();
        if (isprint (c)) {
            if (pos >= len - 1) {
                continue;
            }
            buf[pos++] = c;
            waddch (mLogWin, c);
        } else if ('\n' == c || '\r' == c) {
            break;
        } else if (KEY_BACKSPACE == c && pos) {
            pos--;
            waddstr (mLogWin, "\x8 \x8"); /* Backspace, space, backspace. */
        } else if (27 == c) { /* Escape. */
            pos = 0;
            break;
        }/* else {
            debug.log ("getstr: unhandled char %d", c);
        }*/
    }

    buf[pos] = 0;
    snprintf (msg, 80, "%s %s", prompt, buf);
    msg[79] = 0;
    strcpy (&mLogHistory[savehistidx*80], msg);

    return pos;
}


void
shASCIIInterface::logPrint (const char *str)
{
    wattrset (mLogWin, ColorMap[mColor]);
    if (mLogRow != 0 && !mNoNewline) {
        waddch (mLogWin, '\n');
        mLogCol = 0;
    }
    mNoNewline = 0;
    waddstr (mLogWin, str);
    wattrset (mLogWin, A_NORMAL);
}


int
shASCIIInterface::diag (const char *format, ...)
{
    int res;
    va_list ap;

    mColor = kMagenta;
    va_start (ap, format);
#if 0
    res = vp (format, ap);
#else
    char dbgbuf[100];
    res = vsnprintf (dbgbuf, 100, format, ap);
    if (mDiagWin) {
        wattrset (mDiagWin, ColorMap[kCyan]);
        waddstr (mDiagWin, dbgbuf);
        waddch (mDiagWin, '\n');
        touchwin (mDiagWin);
        wnoutrefresh (mDiagWin);
        doupdate ();
    }
    debug.log ("%s", dbgbuf);
#endif
    va_end (ap);
    mColor = kGray;
    return res;
}


void
shASCIIInterface::pageLog ()
{
    if (Flags.mFadeLog) {
        int x, y;
        getyx (mLogWin, y, x);

        wattrset (mLogWin, ColorMap[kBlue]);

        for (int i = 0; i < mLogRow; i++) {
            int offs = (HISTORY_ROWS + mHistoryIdx - (mLogRow-i)) % HISTORY_ROWS;
            mvwaddstr (mLogWin, i, 0, &mLogHistory[offs*80]);
        }

        wattrset (mLogWin, A_NORMAL);
        wmove (mLogWin, y, x);
        mLogSCount = 0;
    } else {
        werase (mLogWin);
        wmove (mLogWin, 0, 0);
        mLogRow = 0;
        mLogCol = 0;
        mLogSCount = 0;
    }
    drawLog ();
}


void
shASCIIInterface::showHistory ()
{
    WINDOW *win;
    PANEL *panel;
    int i = 0;
    int n = 0;

    win = newwin (25, 80, 0, 0);
    if (!win) {
        p ("Couldn't open history window");
        return;
    }
    panel = new_panel (win);

    mvwaddstr (win, 0, 0, "Console history:");
    if (mHistoryWrapped) {
        i = (HISTORY_ROWS + mHistoryIdx -
             (HISTORY_ROWS < 23 ? HISTORY_ROWS : 23)) % HISTORY_ROWS;
    }
    do {
        mvwaddstr (win, n + 1, 0, &mLogHistory[i*80]);
        i = (i + 1) % HISTORY_ROWS;
        ++n;
    } while (i != mHistoryIdx);
    mvwaddstr (win, n + 1, 0, "--End--");

    update_panels ();
    doupdate ();
    getChar ();
    hide_panel (panel);
    del_panel (panel);
    delwin (win);
    update_panels ();
    drawScreen ();
}


void
shASCIIInterface::showKills ()
{
    int entries = 0, longest = 0;
    for (int i = 0; i < kNumMonIlks; ++i) {
        if (MonIlks[i]->mKills) {
            ++entries;
            int len = strlen (MonIlks[i]->mName);
            longest = maxi (len, longest);
        }
    }
    longest += 15; /* Should suffice. */
    char *lines = (char *) calloc (entries * longest, sizeof (char));
    entries = 0;
    char *buf = GetBuf ();
    for (int i = 0; i < kNumMonIlks; ++i) {
        if (MonIlks[i]->mKills) {
            snprintf (buf, 80, "%s", MonIlks[i]->mName);
            if (MonIlks[i]->mKills > 1) {
                makePlural (buf, 80);
            }
            sprintf (lines + entries * longest, "@%c%c@H - %d %s",
                MonIlks[i]->mGlyph.mColor + 'A', MonIlks[i]->mGlyph.mSym,
                MonIlks[i]->mKills, buf);
            ++entries;
        }
    }
    shTextViewer *kill_list = newViewer (lines, entries, longest);
    kill_list->show ();
}


void
shASCIIInterface::doScreenShot (int fileHandle)
{
    struct shWindata
    {
        int y1, x1, y2, x2;
        WINDOW *winptr;
    };
    shWindata winyx[2] =
    {
        { 0,     0,     mYMax, mXMax, mMainWin },
        { 0,     mXMax, mYMax, 80,    mSideWin }
    };
    chtype square;
    char c;
    char screen[20][81];
    /* Fill screenshot buffer. */
    for (int win = 0; win < 2; ++win) {
        for (int y = winyx[win].y1; y < winyx[win].y2; ++y) {
            for (int x = winyx[win].x1; x < winyx[win].x2; ++x) {
                square = mvwinch (winyx[win].winptr,
                    y - winyx[win].y1, x - winyx[win].x1);
                c = square & A_CHARTEXT;    /* Strip color. */
                screen[y][x] = c;
            }
        }
    } /* Dump everything to file. */
    for (int y = 0; y < 20; ++y) {
        screen[y][80] = '\n';
        write (fileHandle, screen[y], 81);
    } /* Copy last entries of log file. */
    for (int y = mLogSCount - 1; y >= 0; --y) {
        int idx = (mHistoryIdx - y - 1 + HISTORY_ROWS) % HISTORY_ROWS;
        if (idx < 0) {
            idx += HISTORY_ROWS;
        }
        int len = strlen (&mLogHistory[idx * 80]);
        write (fileHandle, &mLogHistory[idx * 80], len);
        write (fileHandle, "\n", 1);
    }
}


void
shASCIIInterface::showVersion ()
{
    const int len = strlen(splash_logo[0]);
    WINDOW *win;
    PANEL *panel;
    win = newwin (18, 60, 3, 10);
    if (!win) {
        return;
    }
    panel = new_panel (win);

    for (int y = 0; y < 8; ++y) {
        for (int x = 0; x < len; ++x) {
            if (splash_logo[y][x] == '#') {
                wattrset (win, ColorMap[kBrightGreen]);
            } else {
                wattrset (win, ColorMap[kGreen]);
            }
            mvwaddnstr (win, y, x + 15, &splash_logo[y][x], 1);
        }
    }

    wattrset (win, ColorMap[kBrown]);
    mvwaddstr (win, 7, 20, "Predator grounds/"PRIME_VERSION"");
    mvwaddstr (win, 9, 8, "Pesky Reticulans, Improvements, More Everything");
    /* The everything part includes bugs.  Tell no one! -- MB */
    wattrset (win, ColorMap[kBrightYellow]);
    mvwaddstr (win, 9, 8,  "P");
    mvwaddstr (win, 9, 14, "R");
    mvwaddstr (win, 9, 26, "I");
    mvwaddstr (win, 9, 40, "M");
    mvwaddstr (win, 9, 45, "E");
    wattrset (win, ColorMap[kBrightRed]);
    mvwaddstr (win, 11, 14, "Unofficial variant of ZAPM v" ZAPM_VERSION " ");
    mvwaddstr (win, 12, 14, "by Psiweapon and Michal Bielinski");
    wattrset (win, ColorMap[kRed]);
    mvwaddstr (win, 13, 13, "http://arcywidmo.republika.pl/prime/");
    wattrset (win, A_NORMAL);
    mvwaddstr (win, 15, 15, "ZAPM (C) 2002-2010 Cyrus Dolph.");
    mvwaddstr (win, 16, 20, "All rights reserved.");
    mvwaddstr (win, 17, 20, "http://www.zapm.org");
    wattrset (win, ColorMap[kBrightGreen]);
    wattrset (win, A_NORMAL);

    update_panels ();
    doupdate ();
    getChar ();
    hide_panel (panel);
    del_panel (panel);
    delwin (win);
    update_panels ();
}
