#include <math.h>
#include <stdio.h>
#include "Global.h"
#include "Util.h"
#include "Monster.h"
#include "Creature.h"
#include "ObjectIlk.h"
#include "Object.h"
#include "Interface.h"
#include "Hero.h"

#define MAXMONSTERLEVEL 25
static shVector <shMonId> MonsterIlksGroupedByLevel[MAXMONSTERLEVEL];

shObjId lizard[NUMLIZARDS];
void addLizardBreaths ();

void
initializeMonsters ()
{
    shMonsterIlk *cur;

#include "MonsterData.h"

    int roll = RNG (3);
    lizard[0] = roll == 0 ? kObjHeatRayGun :
                roll == 1 ? kObjFreezeRayGun :
                            kObjPoisonRayGun;
    roll = RNG (6);
    lizard[1] = roll == 0 ? kObjHeatRayGun :
                roll == 1 ? kObjFreezeRayGun :
                roll == 2 ? kObjPoisonRayGun :
                roll == 3 ? kObjTransporterRayGun :
                roll == 4 ? kObjGammaRayGun :
                            kObjStasisRayGun;
    roll = RNG (6);
    lizard[2] = roll == 0 ? kObjPoisonRayGun :
                roll == 1 ? kObjGammaRayGun :
                roll == 2 ? kObjTransporterRayGun :
                roll == 3 ? kObjGaussRayGun :
                roll == 4 ? kObjStasisRayGun :
                            kObjDisintegrationRayGun;
}

void addLizardBreaths ()
{
    for (int i = 0; i < NUMLIZARDS; ++i) {
        /* Clone and modify ray gun zap effect. */
        shAttack *atk = new shAttack (*(AllIlks[lizard[i]]->mZapAttack));
        shMonsterIlk *mon = NULL;
        switch (i) {
        case 0: mon = MonIlks[kMonLizard1]; break;
        case 1: mon = MonIlks[kMonLizard2]; break;
        case 2: mon = MonIlks[kMonLizard3]; break;
        }
        if (!mon) continue;
        atk->mProb = 5;
        atk->mRange = 8;
        atk->mAttackTime = 2000 - 500 * i;
        atk->mDamage[0].mHigh = atk->mDamage[0].mHigh * (i+1) / 2;
        mon->mRangedAttacks.add (atk);
    }
}

//constructor:
shMonsterIlk::shMonsterIlk (shMonId id,
                            const char *name,
                            shCreatureType ty,
                            shMonsterIlk *parent,
                            shThingSize size,
                            int hitdice,
                            int baselevel,
                            int str, int con, int agi, int dex,
                            int in, int wis, int psi,
                            int speed,
                            int gender,
                            int numhands,
                            int ac, /*natural armor bonus */
                            int numappearingdice,
                            int numappearingdiesides,
                            shMonster::Strategy strategy,
                            int peacefulchance,
                            int probability,
                            char sym,
                            shColor color)
    : mAttacks () , mRangedAttacks ()
{
    mId = id;
    mKills = 0;
    mName = name;
    mType = ty;
    mParent = parent;
    mSize = size;
    mHitDice = hitdice;
    mBaseLevel = baselevel;
    mMutantLevel = 0;
    mStr = str;
    mCon = con;
    mAgi = agi;
    mDex = dex;
    mInt = in;
    mPsi = psi;
    mSpeed = speed;
    mGender = gender;
    mNumHands = numhands;
    mNaturalArmorBonus = ac;
    mInnateIntrinsics = 0;
    for (int i = 0; i < kMaxEnergyType; i++) {
        mInnateResistances[i] = 0;
    }
    mFeats = 0;
    mNumAppearingDice = numappearingdice;
    mNumAppearingDieSides = numappearingdiesides;
    mDefaultStrategy = strategy;
    mDefaultDisposition = shMonster::kHostile;
    mPeacefulChance = peacefulchance;
    mProbability = probability;
    mAttackSum = 0;
    initGlyph (&mGlyph, sym, color);

    for (int i = 0; i < kMaxAllPower; i++) {
        mMutantPowers[i] = 0;
    }

    MonIlks[id] = this;
    MonsterIlksGroupedByLevel[mBaseLevel].add (id);
}


void
shMonsterIlk::addMutantPower (shMutantPower power)
{
    if (kTelepathyPower == power) {
        mInnateIntrinsics |= kTelepathy;
    } else {
        mMutantPowers[power] = 1;
        ++mMutantLevel;
    }
}

void
shMonsterIlk::addAttack (shAttack::Type type, int etype, int numdice,
                         int dicesides, int attacktime,
                         int prob /* = 1 */,
                         int energy2 /* = kNoEnergy */,
                         int numdice2 /* = 0 */,
                         int dicesides2 /* = 6 */ )
{
    int flags = kMelee;
    shAttack *atk;

    if (mAttacks.count ()) {
        flags |= kSecondaryAttack;
    }
    if (shAttack::kTouch == type) {
        flags |= kTouchAttack;
    }

    atk = newMonsterAttack (type, flags, attacktime, prob,
            (shEnergyType) etype, numdice, numdice * dicesides,
            (shEnergyType) energy2, numdice2, numdice2 * dicesides2);

    mAttacks.add (atk);
    mAttackSum += prob;
}


void
shMonsterIlk::addRangedAttack (shAttack::Type type,
                               shAttack::Effect effect,
                               int etype, int numdice,
                               int dicesides, int attacktime,
                               int range, /* = 10 */
                               int prob /* = 1 */,
                               int energy2 /* = kNoEnergy */,
                               int numdice2 /* = 0 */,
                               int dicesides2 /* = 6 */ )
{
    int flags = kAimed;
    shAttack *atk;

    atk = newMonsterRangedAttack (type, effect, flags, attacktime, prob, range,
            (shEnergyType) etype, numdice, numdice * dicesides,
            (shEnergyType) energy2, numdice2, numdice2 * dicesides2);

    mRangedAttacks.add (atk);
}


void
shMonsterIlk::addEquipment (const char *description)
{
    mStartingEquipment.add (description);
}

void
shMonsterIlk::addIntrinsic (shIntrinsics intrinsic)
{
    mInnateIntrinsics |= intrinsic;
}

void
shMonsterIlk::addResistance (shEnergyType energy, int amount)
{
    mInnateResistances[energy] = amount;
}

void
shMonsterIlk::addFeat (shFeat feat)
{
    mFeats |= feat;
}

//constructor:
shMonster::shMonster (shMonId id, int extralevels /* = 0 */ )
    :shCreature ()
{
    int do_or = 0;
    int do_and = 0;
    int gotweapon = 0;
    shMonsterIlk *ilk = MonIlks[id];
    mIlkId = id;

    mType = ilk->mType;
    strncpy (mName, ilk->mName, 30);
    mCLevel = ilk->mBaseLevel;
    mNaturalArmorBonus = ilk->mNaturalArmorBonus;
    mReflexSaveBonus = 0;

    mDir = kNoDirection;
    mTame = 0;
    mStrategy = ilk->mDefaultStrategy;
    mDisposition = ilk->mDefaultDisposition;
    mTactic = kReady;
    mDestX = -1;
    mEnemyX = -1;
    mPlannedMoveIndex = -1;
    mSpellTimer = 0;

    memcpy (mInnateResistances, ilk->mInnateResistances,
            sizeof (mInnateResistances));
    mInnateIntrinsics = ilk->mInnateIntrinsics;
    mFeats = ilk->mFeats;

#define IMMUNE 122

    switch (mType) {
    case kMutant:
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateResistances[kMagnetic] = IMMUNE;
        break;
    case kHumanoid:
        mInnateResistances[kMagnetic] = IMMUNE;
        break;
    case kAnimal:
        mInnateResistances[kMagnetic] = IMMUNE;
        mInnateIntrinsics |= kScent;
        break;
    case kInsect:
        mInnateResistances[kMagnetic] = IMMUNE;
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateIntrinsics |= kScent;
        break;
    case kOutsider:
        mInnateResistances[kMagnetic] = IMMUNE;
        break;
    case kBot:
        mInnateResistances[kPoisonous] = IMMUNE;
        mInnateResistances[kPsychic] = IMMUNE;
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateResistances[kStunning] = IMMUNE;
        mInnateResistances[kHealing] = IMMUNE;
        mInnateResistances[kRestoring] = IMMUNE;
        mInnateResistances[kChoking] = IMMUNE;
        mInnateResistances[kViolating] = IMMUNE;
        mInnateResistances[kParalyzing] = IMMUNE;
        mInnateResistances[kMesmerizing] = IMMUNE;
        mInnateResistances[kBlinding] = IMMUNE;
        mInnateIntrinsics |= kBreathless;
        break;
    case kDroid:
        mInnateResistances[kPoisonous] = IMMUNE;
        mInnateResistances[kPsychic] = IMMUNE;
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateResistances[kStunning] = IMMUNE;
        mInnateResistances[kHealing] = IMMUNE;
        mInnateResistances[kRestoring] = IMMUNE;
        mInnateResistances[kChoking] = IMMUNE;
        mInnateResistances[kViolating] = IMMUNE;
        mInnateResistances[kParalyzing] = IMMUNE;
        mInnateResistances[kMesmerizing] = IMMUNE;
        mInnateResistances[kBlinding] = IMMUNE;
        mInnateIntrinsics |= kBreathless;
        break;
    case kProgram:
        mInnateResistances[kPoisonous] = IMMUNE;
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateResistances[kStunning] = IMMUNE;
        mInnateResistances[kHealing] = IMMUNE;
        mInnateResistances[kRestoring] = IMMUNE;
        mInnateResistances[kChoking] = IMMUNE;
        mInnateResistances[kViolating] = IMMUNE;
        mInnateResistances[kParalyzing] = IMMUNE;
        /* Processes can be put to sleep so no kMesmerizing immunity. */
        mInnateIntrinsics |= kBreathless;
        break;
    case kConstruct:
        mInnateResistances[kPoisonous] = IMMUNE;
        mInnateResistances[kPsychic] = IMMUNE;
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateResistances[kStunning] = IMMUNE;
        mInnateResistances[kChoking] = IMMUNE;
        mInnateResistances[kViolating] = IMMUNE;
        mInnateResistances[kParalyzing] = IMMUNE;
        mInnateResistances[kMesmerizing] = IMMUNE;
        mInnateResistances[kBlinding] = IMMUNE;
        mInnateIntrinsics |= kBreathless;
        break;
    case kOoze:
        mInnateResistances[kPoisonous] = 3;
        mInnateResistances[kMagnetic] = IMMUNE;
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateResistances[kChoking] = IMMUNE;
        mInnateResistances[kViolating] = IMMUNE;
        mInnateIntrinsics |= kBreathless;
        break;
    case kAberration:
        mInnateResistances[kMagnetic] = IMMUNE;
        mInnateResistances[kRadiological] = IMMUNE;
        mInnateResistances[kViolating] = IMMUNE;
        break;
    case kCyborg:
        mInnateResistances[kPoisonous] = 2;
        mInnateResistances[kRadiological] = 10;
        mInnateIntrinsics |= kBreathless;
        break;
    case kEgg:
        mInnateResistances[kViolating] = IMMUNE;
        mInnateIntrinsics |= kBreathless;
        /* fall through */
    case kBeast:
        mInnateResistances[kMagnetic] = IMMUNE;
        mInnateIntrinsics |= kScent;
        break;
    case kVermin:
        mInnateResistances[kMagnetic] = IMMUNE;
        mInnateIntrinsics |= kScent;
        mInnateIntrinsics |= kCanSwim;
        break;
    case kAlien:
        mAlienEgg.mHatchChance = 0;
        mInnateResistances[kMagnetic] = IMMUNE;
        mInnateResistances[kRadiological] = 10;
        mInnateResistances[kCorrosive] = IMMUNE;
        mInnateIntrinsics |= kScent;
        gainRank (kUnarmedCombat, mCLevel/2);
        break;
    default:
        debug.log ("Alert! Unknown monster type");
        mInnateResistances[kMagnetic] = IMMUNE;
    }

#undef IMMUNE

    if (3 == ilk->mGender) {
        mGender = RNG (1, 2);
    } else {
        mGender = ilk->mGender;
    }

    /* roll ability scores */
    rollAbilityScores (ilk->mStr, ilk->mCon, ilk->mAgi, ilk->mDex,
                       ilk->mInt, ilk->mPsi);

    /* roll hit points */
    rollHitPoints (ilk->mHitDice, 8);

    /* setup speed */
    mSpeed = ilk->mSpeed;
    computeAC ();

    mGlyph = ilk->mGlyph;
    mState = kActing;

    for (int i = 0; i < ilk->mStartingEquipment.count (); i++) {
        const char *str = ilk->mStartingEquipment.get (i);

        if ('|' == str[0]) { /* the or symbol indicates don't try to create
                                the object unless the previous obj failed */
            if (0 == do_or) {
                do_or = 1;
                do_and = 0;
                continue;
            } else {
                ++str;
            }
        } else if (',' == str[0]) { /* the comma symbol indicates create the
                                       object only if the prev obj was made */
            if (1 == do_and) {
                ++str;
            } else {
                continue;
            }
        }

        shObject *obj = createObject (str, 0);

        if (NULL == obj) {
            do_or = 1;
            do_and = 0;
            debug.log ("unable to equip %s", str);
            continue;
        }
        do_or = 0;
        do_and = 1;
        addObjectToInventory (obj);
        if (!mWeapon && obj->isA (kWeapon)) {
            if (obj->myIlk ()->mMeleeSkill != kNoSkillCode) {
                gainRank (obj->myIlk ()->mMeleeSkill, 1 + mCLevel * 2/3);
            }
            if (obj->myIlk ()->mGunSkill != kNoSkillCode) {
                gainRank (obj->myIlk ()->mGunSkill, 1 + mCLevel * 2/3);
            }
            ++gotweapon;
            /* don't wield until hero is in sight so he can see message */
            //wield (obj, 1);
        } else if (obj->isA (kArmor) || obj->isA (kImplant)) {
            don (obj, 1);
        }
    }

    if (!gotweapon) {
        gainRank (kUnarmedCombat, 1 + mCLevel);
    }

    /* Maybe monster gets some more treasure. */
    if (noTreasure () ||
        kAnimal == mType  || kBeast == mType    || kAberration == mType ||
        kInsect == mType  || kVermin == mType   || kOutsider == mType ||
        kEgg == mType     || kOoze == mType     || kAlien == mType ||
        kBot == mType     || kConstruct == mType)
    { /* No treasure requested or a monster does not want treasure. */
    } else {
        if (RNG (50) <= 5 + mCLevel) {
            shObject *cash = new shObject (kObjMoney);
            cash->mCount = NDX (mCLevel + 1, 10);
            addObjectToInventory (cash);
        }
        if (RNG (80) <= 5 + mCLevel) {
            shObject *obj = generateObject (-1);

            while (obj->getMass () > 5000) {
                delete obj;
                obj = generateObject (-1);
            }

            addObjectToInventory (obj);
        }
    }

    computeIntrinsics ();

    if (RNG (100) < ilk->mPeacefulChance) {
        mDisposition = kIndifferent;
    }

    debug.log ("spawned %s with %d HP speed %d", mName, mHP, mSpeed);
}


shMonId
pickAMonsterIlk (int level)
{
    int baselevel = level;
    int n = 0;
    shVector <shMonId> v;
    shMonId ilk;
    int i;
    int c;

//    rarity = RNG (3) ? 1 : RNG (3) ? 2 : 3;

    while (0 == (n = MonsterIlksGroupedByLevel[baselevel].count ())) {
        baselevel--;
    }


    for (i = 0; i < n; i++) {
        ilk = MonsterIlksGroupedByLevel[baselevel].get (i);
        c = MonIlks[ilk]->mProbability;
        while (c--) {
            v.add (ilk);
        }
    }
    if (0 == v.count ()) {
        return kMonEarthling;
    }

    return v.get (RNG (v.count ()));
}


shMonster *
generateMonster (int level)
{
    /* generates random monster of the given level */

    int baselevel = level;
    int n;

    while (0 == (n = MonsterIlksGroupedByLevel[baselevel].count ())) {
        baselevel--;
    }

    return new shMonster (MonsterIlksGroupedByLevel[baselevel].get (RNG (n)),
                          baselevel - level);
}


const char *
shMonster::the ()
{
    char *buf = GetBuf ();

    if (!Hero.canSee (this) && !Hero.canHearThoughts (this)) {
        strncpy (buf, hasMind () ? "someone" : "something", SHBUFLEN);
    } else if (Hero.is (kXenosHunter) && isXenos ()) {
        snprintf (buf, SHBUFLEN, "the xenos scum");
    } else if (Hero.is (kXenosHunter) && isHeretic ()) {
        snprintf (buf, SHBUFLEN, "the heretic");
    } else {
        snprintf (buf, SHBUFLEN, "the %s", myIlk ()->mName);
    }
    return buf;
}

const char *
shMonster::an ()
{
    char *buf = GetBuf ();

    if (!Hero.canSee (this) && !Hero.canHearThoughts (this)) {
        strncpy (buf, hasMind () ? "someone" : "something", SHBUFLEN);
    } else if (isUnique ()) {
        return the ();
    } else if (Hero.is (kXenosHunter) && isXenos ()) {
        snprintf (buf, SHBUFLEN, "a xenos scum");
    } else if (Hero.is (kXenosHunter) && isHeretic ()) {
        snprintf (buf, SHBUFLEN, "a heretic");
    } else {
        snprintf (buf, SHBUFLEN,
                  isvowel (myIlk ()->mName[0]) ? "an %s" : "a %s",
                  myIlk ()->mName);
    }
    return buf;
}

const char *
shMonster::your ()
{
    char *buf = GetBuf ();

    if (Hero.isBlind () && !Hero.canHearThoughts (this)) {
        strncpy (buf, hasMind () ? "someone" : "something", SHBUFLEN);
    } else {
        snprintf (buf, SHBUFLEN, "your %s", myIlk ()->mName);
    }
    return buf;
}


const char *
shMonster::getDescription ()
{
    return myIlk ()->mName;
}

int
shMonster::numHands ()
{
    return myIlk ()->mNumHands;
}

int
shMonster::die (shCauseOfDeath how, const char *killer)
{
    return shCreature::die (how, killer);
}

/*     SPOILER GENERATION     */

int speedToAP (int speed); /* Game.cpp */

void
shMonsterIlk::spoilers (FILE *spoilerfile)
{
    int i, n = 100;
    int AC = 0;
    int speed = 0;
    int HP = 0;
    int ap = 0;

    int acfoo = 4 * mBaseLevel / 3  + 10;

    int attacks = 0;
    int hits = 0;
    int damagefoo = 0;
    int damage15 = 0;
    int damage20 = 0;
    int damage25 = 0;

    int rhits = 0;
    int rattacks = 0;
    int rdamagefoo = 0;
    int zhits = 0;

    shMonster *m;

    for (i = 0; i < n; i++) {
        m = new shMonster (mId);
        int clk = 0;

        AC += m->getAC ();
        speed += m->mSpeed;
        HP += m->mMaxHP;

        m->readyWeapon ();

        if (m->mWeapon && m->mWeapon->isA (kRayGun)) {
            zhits += 1;
            continue;
        }

        while (clk < 60000) {
            shAttack *atk = NULL;

            int attack = 0;
            int damage = 0;

            if (!mAttacks.count ()) {
                break;
            }

            clk += 100;
            ap += speedToAP (m->mSpeed);
            if (ap < 0) {
                continue;
            }

            attack += RNG (1, 20);
            if (20 == attack) attack = 100;

            if (m->mWeapon) {
                if (m->mWeapon->isMeleeWeapon ()) {
                    atk = m->mWeapon->myIlk ()->mMeleeAttack;
                    damage = ABILITY_MODIFIER (m->getStr ());
                } else if (m->mWeapon->isThrownWeapon ()) {
                    atk = m->mWeapon->myIlk ()->mMissileAttack;
                    damage = ABILITY_MODIFIER (m->getStr ());
                } else {
                    atk = m->mWeapon->myIlk ()->mGunAttack;
                    damage = 0;
                    --attack; /* assume range penalty */
                }
                attack += m->mWeapon->myIlk ()->mToHitBonus;
                attack += m->getWeaponSkillModifier (m->mWeapon->myIlk (), atk);
                attack += m->mWeapon->mEnhancement;
                damage += m->mWeapon->mEnhancement;
                if (!m->mWeapon->isMeleeWeapon ()) {
                    rattacks++;
                } else {
                    attacks++;
                }
            } else {
                atk = mAttacks.get (RNG (mAttacks.count ()));
                if (atk->isSecondaryAttack ()) {
                    damage = ABILITY_MODIFIER (m->getStr ());
                    if (damage > 0) {
                        damage /= 2;
                    }
                } else {
                    damage = ABILITY_MODIFIER (m->getStr ());
                }
                ++attacks;
            }

            ap -= atk->mAttackTime;


            attack += m->mToHitModifier;
            damage += RNG (atk->mDamage[0].mLow,
                           atk->mDamage[0].mHigh);
            if (acfoo > 22) { /* spot hero +2 agi bonus */
                damage -= (acfoo - 21) / 2;
            }
            if (damage < 1) damage = 1;
            if (attack > acfoo) {
                if (m->mWeapon && !m->mWeapon->isMeleeWeapon ()) {
                    rdamagefoo += damage;
                    ++rhits;
                } else {
                    damagefoo += damage;
                    ++hits;
                }

            }
            if (attack >= 15) {
                damage15 += damage;
            }
            if (attack >= 20) {
                damage20 += damage;
            }
            if (attack >= 25) {
                damage25 += damage;
            }
        }

        delete m;
    }



    fprintf (spoilerfile,
             "%28s %2d %3d %4d %4d %3d/%3d %4d %3d/%3d %4d %2d\n",
             mName,
             AC / n, HP / n, speed / n,
             acfoo,
             rhits / (n),
             rattacks / (n),
             rdamagefoo / (n),
             hits / (n),
             attacks / (n),
             damagefoo / (n),
             zhits);

/*
    I->p ("                                     %4d/15 %4d/20 %4d/25",
          damage15 / (6 * n),
          damage20 / (6 * n),
          damage25 / (6 * n));
*/
}





void
monsterSpoilers ()
{
    FILE *spoilerfile = fopen ("monsters.txt", "w");

    I->p ("Generating spoilers, please be patient...");

    fprintf (spoilerfile, "%28s %2s %3s %4s %2s %3s/%3s %4s %3s/%3s %4s\n",
             "Monster", "AC", "HP", "Spd", "vsAC", "hit", "rgd", "dmg", "hit", "atk", "dmg");

    for (int i = 0; i < kNumMonIlks; ++i) {
        shMonsterIlk *ilk = MonIlks[i];
        ilk->spoilers (spoilerfile);
    }
    fclose (spoilerfile);
    I->p ("Spoilers saved in monsters.txt.");
}
